########################################################################
########################################################################
###
###	Filename: PriorityQueue.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
########################################################################
########################################################################
###
###	Description: This file implements a constructor `PriorityQueue'
###	for priority queue objects. The implementation uses a heap to
###	maintain a partially ordered balanced tree.
###
###	SYNOPSIS
###
###		PriorityQueue( priority::procedure [,rest] )::priqueue
###
###	PARAMETERS
###
###		priority -- given an item on the queue, real-valued
###			    procedure that returns its priority
###
###	DESCRIPTION
###
###	This procedure returns a priority queue object, represented as a
###	module in which a heap data structure is used to maintain the
###	queue. The first (required) argument is a procedure PRIORITY that
###	computes the priority of any item that may be placed upon the
###	queue. This procedure must be real-valued (numeric), so that a
###	comparison that uses "<" or "<=" will succeed, and should not
###	return an error.
###
###	The priority queue object is returned. It exports four methods:
###
###		init -- initialize the priority queue
###		empty -- test whether the queue is empty
###		top -- return (non-destructively) the item at the top
###			of the queue
###		insert -- put an item onto the queue
###		delete -- remove the highest priority item from the queue
###
########################################################################
########################################################################

# Change this to be $define'd if you want to use the (tail) recursive
# forms of the local routines BUBBLEUP and BUBBLEDOWN.
$undef	RECURSIVE

`type/priqueue` := '`module`( top, insert, delete, empty )':
#savelib( '`type/priqueue`' ):

unprotect( 'PriorityQueue' ):
PriorityQueue	:=	proc( priority::procedure )
	local	largs, lnargs;

	# Save any argument to be queued in `largs'
	lnargs := nargs;
	if lnargs > 1 then
		largs := [ args[ 2 .. -1 ] ]
	else
		largs := []
	end if;

	# Construct the priority queue object
	module()

		description "a priority queue";

		export	empty,		# test for an empty queue
			top,		# return the top item on the queue
			insert,		# insert an item onto the queue
			delete,		# remove the top item on the queue
			size,		# return the number of items on the queue
			init;		# reinitialize the priority queue
		local	heap,		# underlying heap data structure
			nitems,		# count of the number of items on the queue
			bubbledown,	# internal routine for maintaining the heap
					# property of HEAP
			bubbleup;	# internal routine for maintaining the heap
					# property of HEAP

		#######################
		# PRIVATE DEFINITIONS #
		#######################

		# The heap starts out empty
		nitems := 0;

		# The heap is maintained in this table, which is actually
		# a growable array. The topmost item on the heap is kept
		# in HEAP[ 1 ]. For a node in HEAP[ n ], the left child is
		# in HEAP[ 2 * n ] and the right child is in HEAP[ 1 + 2 * n ].
		# Thus, the heap looks like:
		#           		HEAP[ 1 ]
		#		HEAP[ 2 ]	HEAP[ 3 ]
		# 	HEAP[ 4 ]  HEAP[ 5 ] HEAP[ 6 ] HEAP[ 7 ]
		#			.......

		heap := table();

		# Local method used to maintain the heap property
		# of `heap'.
$ifdef	RECURSIVE
		bubbleup := proc( child::posint )
			local	parent;
			parent := iquo( child, 2 );
			if child > 1
			    and priority( heap[ child ] ) > priority( heap[ parent ] ) then
				heap[ parent ], heap[ child ] := heap[ child ], heap[ parent ];
				procname( parent )
			end if
		end proc;
$else	# not RECURSIVE
		bubbleup := proc( child::posint )
			local	parent, c;
			c := child;
			while c > 1 do
				parent := iquo( c, 2 );
				if priority( heap[ c ] ) >= priority( heap[ parent ] ) then
					heap[ c ], heap[ parent ] := heap[ parent ], heap[ c ];
					c := parent
				else
					break
				end if
			end do
		end proc;
$endif	# RECURSIVE

$ifdef	RECURSIVE
		bubbledown := proc( parent::posint )
			local	child;
			child := 2 * parent;
			if child < nitems
			    and priority( heap[ 1 + child ] ) > priority( heap[ child ] ) then
				child := 1 + child
			end if;
			if child <= nitems and priority( heap[ parent ] ) < priority( heap[ child ] ) then
				heap[ parent ], heap[ child ] := heap[ child ], heap[ parent ];
				procname( child ) # new parent
			end if
		end proc;
$else	# not RECURSIVE
		bubbledown := proc( parent::posint )
			local	p, child;

			p := parent;
			while 2 * p <= nitems do
				child := 2 * p;
				if child < nitems
				    and priority( heap[ child ] ) < priority( heap[ 1 + child ] ) then
					# pick the larger of the two children
					child := 1 + child
				end if;
				if priority( heap[ p ] ) < priority( heap[ child ] ) then
					heap[ p ], heap[ child ] := heap[ child ], heap[ p ];
					p := child
				else
					break
				end if
			end do
		end proc;
$endif	# RECURSIVE

		####################
		# EXPORTED METHODS #
		####################

		# Reinitialize the priority queue.
		init := proc()
			heap := table();
			nitems := 0
		end proc;

		# Test whetheer the queue is empty
		empty := () -> evalb( nitems < 1 );

		# Return the number of items on the queue.
		size := () -> nitems;

		# Return the item at the top of the queue (that which has the
		# highest priority).
		top := proc()
			if empty() then
				error "priority queue is empty"
			else
				heap[ 1 ]
			end if
		end proc;

		# Remove the highest priority item from the queue.
		delete := proc()
			local	val;
			val := heap[ 1 ]; # val := top()
			# move bottom to the top
			heap[ 1 ] := heap[ nitems ];
			# allow expression to be collected
			heap[ nitems ] := evaln( heap[ nitems ] );
			# decrement the bottom of heap pointer
			nitems := nitems - 1;
			# heapify the array
			bubbledown( 1 );
			# return the value
			val
		end proc;

		# Insert an item onto the queue.
		insert := proc( v )
			if nargs > 1 then
				return map( procname, [ args ] )
			end if;
			nitems := 1 + nitems;
			heap[ nitems ] := v;
			bubbleup( nitems )
		end proc;

		###################
		# INITIALIZATIONS #
		###################

		# Initiate the priority queue with any given items
		if lnargs > 1 then
			insert( op( largs ) )
		end if

	end module
end proc:
protect( 'PriorityQueue' ):
#savelib( 'PriorityQueue' ):

########################################################################
########################################################################
###
###	End of File: PriorityQueue.mpl
###
########################################################################
########################################################################
