########################################################################
########################################################################
###
###	Filename: QuotientField.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
########################################################################
########################################################################
###
###	Description: This file contains implementations of a generic
###	constructor (functor) `QuotientField()'.
###
###	The functor `QuotientField()' constructs the quotient field (or
###	``field of fractions'') of an integral domain.
###
###	For computations, we restrict ourselves to rings which we may
###	suppose are equipped with division with remainder and greatest
###	common divisors.
###
########################################################################
########################################################################

# This is an utility type that allows us to represent procedures
# as compositions of other procedures. Note that this is a recursively
# defined type.
`type/Procedure` := '{ procedure, specfunc( Procedure, `@` ) }':
#savelib( '`type/Procedure`' ):

# This type describes the rings we are working with.
`type/Ring` := '`module`(
	# Arithemetic operators
	`+`	::	Procedure,
	`*`	::	Procedure,
	`-`	::	Procedure,
	# Constants (0/1)
	zero,	#	untyped
	one,	#	untyped
	# Tests (for 0/1)
	iszero	::	Procedure,
	isone	::	Procedure,
	# Division
	quo	::	Procedure,
	rem	::	Procedure,
	# Greatest common divisor
	gcd	::	Procedure,
	# Unit normal form
	unormal	::	Procedure,
	# Embed the integers
	zembed	::	procedure(integer)
)':
#savelib( '`type/Ring`' ):

# This is a description of a field.
`type/Field` := '`module`(
	# Arithemetic operators
	`+`	::	Procedure,
	`*`	::	Procedure,
	`-`	::	Procedure,
	`/`	::	Procedure,
	# Constants (0/1)
	zero,	#	untyped
	one,	#	untyped
	# Tests (for 0/1)
	iszero	::	Procedure,
	isone	::	Procedure,
	# Field element constructor
	make	::	Procedure,
	# Normal form
	normal	::	Procedure,
	# Embedding of base ring
	embed	::	Procedure,
	# Embedding of ring of integers
	zembed	::	procedure(integer)
)':
#savelib( '`type/Field`' ):

# An ideal in a ring can be described using this type.
`type/Ideal` := '`module`(
	# Pointer to the parent ring
	ring	::	Ring,
	# An effective test for membership in the ideal
	member	::	Procedure
)':
#savelib( '`type/Ideal`' ):

# This is the constructor for quotient fields.
QuotientField := proc( D::Ring )
	description "quotient field functor";
	option	inline;
	module()
		description "a quotient field";
		export `+`, `*`, `-`, `/`,	# arithmetic operators
			make,			# constructor
			numer, denom,		# syntax routines
			zero, one,		# constants
			iszero, isone,		# tests
			normal,			# normal form
			embed,			# embed D
			zembed;			# embed the integers

		# Construct a fraction from its numerator
		# and denominator.
		make := proc( n, d )
			local	u, nd;
			if D:-iszero( d ) then
				error "division by zero"
			end if;
			u, nd := D:-unormal( d );
			'FRACTION'( u * n, nd )
		end proc;

		# Retrieve the numerator of a fraction.
		numer := proc( f )
			option	inline;
			op( 1, f )
		end proc;

		# Retrieve the denominator of a fraction.
		denom := proc( f )
			option inline;
			op( 2, f )
		end proc;

		# The zero of the field.
		zero := embed( D:-zero );
		# The multiplicative identity of the field.
		one := embed( D:-one );

		# Embed D into the quotient field.
		embed := d -> make( d, D:-one );

		# Embed the ring of integers into the quotient field.
		zembed := i -> embed( D:-zembed( i ) );

		# Tests
		iszero := D:-iszero @ numer @ normal;
		isone := D:-isone @ numer @ normal;

		# Normal form for field elements.
		normal := proc( f )
			local g, a, b;
			# We assume that g = D:-unormal( g ); i.e.,
			# that g is already unit normal
			g := D:-gcd( numer( f ), denom( f ) );
			if D:-isone( g ) then
				f
			else
				a := D:-quo( numer( f ), g );
				b := D:-quo( denom( f ), g );
				make( a, b )
			end if
		end proc;

		# Arithmetic
		`-` := f -> normal( D:-`-`( numer( f ) ), denom( f ) );
		`/` := f -> normal( make( denom( f ), numer( f ) ) );

		`+` := proc( a, b )
			use `+` = D:-`+`, `*` = D:-`*` in
				normal( make( numer( a ) * denom( b ) + denom( a ) * numer( b ),
				              denom( a ) * denom( b ) ) )
			end use
		end proc;

		`*` := proc( a, b )
			use `*` = D:-`*` in
				normal( make( numer( a ) * numer( b ),
				              denom( a ) * denom( b ) ) )
			end use
		end proc;
	end module
end proc:
#savelib( 'QuotientField' ):

########################################################################
########################################################################
###
###	End of File: QuotientField.mpl
###
########################################################################
########################################################################
