################################################################################
################################################################################
###
###	Filename: ShapeObjects.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
################################################################################
################################################################################
###
###	Description: This file contains the source code for the Maple package
###	`ShapeObjects'. This is an "object-oriented" variant of the `Shapes' package.
###	It uses an object representation of shapes instead of representing
###	shapes as unevaluated function calls.
###
###	The package defines a collection of constructors for shape objects.
###	The shapes supported by the package, and the names of their constructors,
###	are:
###
###		point, segment, circle, square, triangle.
###
###	Each returns a shape object that exports, at least, the methods `area'
###	and `circumference'.  Particular shapes may export additional methods.
###
################################################################################
################################################################################

$include	<assert.mi>

ShapeObjects := module()
	description "a demonstration package of shape constructors";
	export	point,
		segment,
		circle,
		square,
		triangle;

	# Constructor for point objects.
	point := proc( x, y )
		description "point constructor";
		# Using option remember ensures that any two points
		# constructed with the same coordinates are equal.
		option	remember, system;

		# The return value is the module instantiated by
		# evaluating the following module definition. It
		# represents a point object. The "state" of the
		# object is stored in the lexically scoped parameters
		# `x' and `y'. The behavior of the object is
		# represented by the procedures defined in the
		# module definition.
		module()
			description "a point";
			export	xcoord, ycoord, area, circumference;

			xcoord := () -> x;
			ycoord := () -> y;
			area := () -> 0;
			circumference := () -> 0;
		end module
	end proc;

	# Constructor for segment objects.
	segment := proc( src, targ )
		description "segment constructor";
		option	remember, system; # uniquely instantiated

		# Error checking.
		if not type( src, '`module`( xcoord, ycoord )' ) then
			error "first argument must be a point object"
		end if;
		if not type( targ, '`module`( xcoord, ycoord )' ) then
			error "second argument must be a point object"
		end if;

		# Instantiate the segment object.
		module()
			description "a segment";
			export	len, midpoint,
				start_point, end_point,
				area, circumference;

			# Return the initial point of the segment.
			start_point := () -> src;
			# Return the terminal point of the segment.
			end_point := () -> targ;
			# Compute the area (0) of the segment.
			area := () -> 0;
			# Compute the circumference (0) of the segment.
			circumference := () -> 0;

			# Compute the length of the segment.
			len := proc()
				local	a1, a2, b1, b2;
				a1 := src:-xcoord();
				a2 := src:-ycoord();
				b1 := targ:-xcoord();
				b2 := targ:-ycoord();
				radnormal( sqrt( ( b2 - a2 )^2 + ( b1 - a1 )^2 ) )
			end proc;

			# Construct the midpoint of the segment.
			midpoint := proc()
				local	a1, a2, b1, b2;
				a1 := src:-xcoord();
				a2 := src:-ycoord();
				b1 := targ:-xcoord();
				b2 := targ:-ycoord();
				point( ( b1 - a1 ) / 2, ( b2 - a2 ) / 2 )
			end proc;
		end module
	end proc;

	# Circle constructor.
	circle := proc( cntr, rad )
		description "circle constructor";
		option	remember, system; # uniquely instantiated

		if not type( cntr, '`module`( xcoord, ycoord )' ) then
			error "first argument (the center) must be a point object"
		end if;

		# Instantiate the circle object.
		module()
			description "a circle";
			export	center, radius, diameter, area, circumference;

			center := () -> cntr;
			radius := () -> rad;
			diameter := () -> 2 * radius();
			area := () -> Pi * radius()^2;
			circumference := () -> Pi * diameter();
		end module
	end proc;

	# Constructor for square objects.
	square := proc( cntr, side )
		description "square constructor";
		option	remember, system; # uniquely instantiated

		if not type( cntr, '`module`( xcoord, ycoord )' ) then
			error "first argument (the center) must be a point object"
		end if;
		if type( side, 'complex( numeric )' ) and not type( side, 'positive' ) then
			error "numeric side lengths must be positive"
		end if;

		# Instantiate the square object.
		module()
			description "a square";
			export	center, side_length, area, circumference;

			# These two routines are simply syntax procedures
			# that return the parts of a square.
			center := () -> cntr;
			side_length := () -> side;
			# Compute the area of the square.
			area := () -> side^2;
			# Compute the circumference of the square.
			circumference := () -> 4 * side;
		end module
	end proc;

	# Triangle constructor.
	triangle := proc( v1, v2, v3 )
		description "triangle constructor";
		option	remember, system; # uniquely instantiated

		if nargs <> 3 then
			error "three vertices are required"
		end if;
		if map( type, [ args ], '`module`( xcoord, ycoord )' ) <> [ true, true, true ] then
			error "arguments must be point objects"
		end if;

		# Instantiate the triangle object. Rather than store
		# the vertices, the line segments that form the sides
		# of the triangle are stored.
		module()
			description "a triangle";
			local	s1, s2, s3;
			export	side, area, circumference;

			# Construct the sides of the triangle.
			s1 := segment( v1, v2 );
			s2 := segment( v2, v3 );
			s3 := segment( v3, v1 );

			# These assertions guard against `segment' failing to
			# produce a segment (for example, by returning unevaluated)
			# and against changes to the segment interface.
			Assert( type( s1, '`module`( midpoint, end_point, len )' ) );
			Assert( type( s2, '`module`( midpoint, end_point, len )' ) );
			Assert( type( s3, '`module`( midpoint, end_point, len )' ) );

			# Return one side of the triangle.
			side := ( n::{1,2,3} ) -> [ s1, s2, s3 ][ n ];

			# Compute the area of the triangle by using the formula
			# area = base * height / 2.
			area := proc()
				local	base, height;
				base := s1:-len();
				height := segment( s1:-midpoint(), s2:-end_point() ):-len();
				Normalizer( base * height / 2 )
			end proc;

			# Compute the circumference of the triangle by summing the
			# lengths of its sides.
			circumference := proc()
				local	i;
				add( side( i ):-len(), i = 1 .. 3 )
			end proc;
		end module
	end proc;
end module:

# If the Maple preprocessor macro `SAVELIBNAME' is defined, then use
# it to define the global variable `savelibname', which determines
# which repository the macro is saved to.
$ifdef	SAVELIBNAME
savelibname := SAVELIBNAME:
$endif	# SAVELIBNAME

# Two methods are used to help automate the loading of this file.
#savelib( 'ShapeObjects' ):
$ifdef	LOAD_REPOSITORY
savelib( 'ShapeObjects' ):
$endif	# LOAD_REPOSITORY

################################################################################
################################################################################
###
###	End of File: ShapeObjects.mpl
###
################################################################################
################################################################################
