################################################################################
################################################################################
###
###	Filename: coverage.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
################################################################################
################################################################################
###
###	Description: This file implements the `coverage' package. This package
###	contains only two exported routines:
###
###		profile -- instruments a procedure or module for coverage
###				profiling
###		covered -- determines which procedures have had profiling data
###				stored
###
################################################################################
################################################################################

# Macros for userinfo messages. Userinfo levels according to
# Maple library standards.
macro( UI_TRIGGER = 'coverage' ):
macro( UI_INFO = 1 ):
macro( UI_TRACE = 3 ):
macro( UI_DEBUG = 5 ):

unprotect( 'coverage' ):
coverage := module()
	description "a package of coverage-profiling utilities";
	option package;
	export	profile,
		covered;
	local	T, plural, rprofile, traced, clear, keys, same_external_rep;

	# This table is used to keep track of which expressions have been
	# profiled, so that an infinite recursion does not occur. Here is
	# an example that would cause an infinite recursion:
	# m := module()
	#     export s;
	#     local me;
	#     me := thismodule;
	#     s := module()
	#         export e;
 	#         e := eval( me, 1 );
	#     end module;
	# end module:
	T := table();

	# Utility for producing a potential plural suffix based
	# on a number `n'.
	plural := proc( n, suffix )
		option	inline;
		`if`( n = 1, "", suffix )
	end proc;

	# Instrument each argument, which will be either a procedure or
	# a module name, for coverage profiling.
	profile := proc()
		description "instrument a procedure or module for coverage profiling";
		local	a, count;
		count := add( rprofile( a ), a = [ args ] );
		userinfo( UI_INFO, UI_TRIGGER,
			sprintf( "%d expression%s instrumented",
				count, plural( count, "s" ) ) );
		NULL
	end proc;

	# Return simple keys for a table.
	keys := proc( t::table )
		option	inline;
		map( op, { indices( t ) } )
	end proc;

	# Return `true' if names `n1' and `n2' have the same
	# external representation; return `false' otherwise.
	same_external_rep := proc( n1, n2 )
		option	inline;
		evalb( convert( n1, 'string' ) = convert( n2, 'string' ) )
	end proc;

	# Remove all entries from the table T that have the same
	# external representation as `s', except `s' itself.
	clear := proc( s )
		local	t;
		for t in select( same_external_rep, keys( T ), s ) minus { s } do
			T[ t ] := evaln( T[ t ] )
		end do
	end proc;

	# Recursively instrument procedures and modules. No recursion
	# occurs for procedures, but modules are traversed recursively.
	rprofile := proc( s::name )
		local	e;

		if type( s, 'procedure' ) then
			# Procedures: basis for the recursion.
			if assigned( T[ s ] ) and T[ s ] = true then
				# already instrumented
				return 0
			end if;
			T[ s ] := true; # mark as done
			userinfo( UI_TRACE, UI_TRIGGER,
				"instrumenting procedure:", s );
			if type( `TRACE/` || s, 'rtable' ) then
				# Another procedure instrumented has a name with
				# the same external representation as s. Remove
				# it and issue a warning.
				WARNING( "overwriting profiling data for %1", s );
				clear( s );
				unassign( evaln( `TRACE/` || s ) )
			end if;
			# Instrument the procedure
			debugopts( 'traceproc' = s );
			# One procedure done, so return 1
			1
		elif type( s, '`module`' ) then
			# Modules: recursive case

			# Check whether this module has been instrumented.
			e := addressof( eval( s ) );
			if assigned( T[ e ] ) and T[ e ] = true then
				return 0
			end if;
			# Not done before, so mark as done now.
			T[ e ] := true;
			userinfo( UI_TRACE, UI_TRIGGER,
				"instrumenting module:", s );
			# Recurse on the module and procedure exports.
			add( procname( e ),
				e = select( type,
					[ exports( s, 'instance' ) ],
					'{`module`,procedure}' ) )
		else
			# Argument type error.
			error "only procedures and modules can be "
				"profiled for coverage"
		end if
	end proc;

	# Test whether a name begins with the substring "TRACE/"
	traced := proc( s::name )
		option	inline;
		evalb( substring( convert( s, 'string' ), 1 .. 6 ) = "TRACE/" )
	end proc;

	# Return those procedures that have been instrumented for coverage
	# profiling. If the optional argument 'nonzero' is passed, only
	# return those for which nonzero data has been collected.
	covered := proc()
		description "query the procedures for which profiling "
			"data has been collected";
		local	S;
		# Find the rtables that store the profiling data.
		S := select( traced, [ anames( 'rtable' ) ] );
		if nargs > 0 and args[ 1 ] = 'nonzero' then
			# pick out those with nonzero coverage
			S := select( s -> evalb( s[1,1] <> 0 ), S )
		elif nargs > 0 then
			error "optional argument is the name \"nonzero\"";
		else
			Assert( nargs = 0 );
		end if;
		# Convert names to the procedures.
		map( parse,
		     map( substring,
			  map( convert, S, 'string' ), 7 .. -1 ) )
	end proc;

end module:
protect( 'coverage' ):
#savelib( 'coverage' ):
$ifdef	LOAD_REPOSITORY
savelib( 'coverage' ):
$endif	# LOAD_REPOSITORY

################################################################################
################################################################################
###
###	End of File: coverage.mpl
###
################################################################################
################################################################################
