########################################################################
########################################################################
###
###	Filename: DiffImpl.mpl
###
########################################################################
########################################################################
###
###	Description: This file contains the implementation of the
###	symbolic differentiator discussed in Chapter 6 of the Maple
###	Programming Guide.
###
###	The differentiator is implemented in a module `DiffImpl', whose
###	principal export is `differentiate'. This export is also assigned
###	to the global name `differentiate' for user access.
###
###	In addition to `differentiate', the module `DiffImpl' exports
###	the following utilities:
###
###		rule	-- display a "rule"
###		addRule	-- add a new "rule"
###		addFunc	-- add the derivative of a new function
###
###	These are used to query and modify the internal database of
###	known derivatives of functions and rules for the differentiation
###	of expressions based on their types. Several examples of their
###	use appear near the end of the body of `DiffImpl'.
###
########################################################################
########################################################################


$include	<assert.mi>

DiffImpl := module()
	description "a symbolic differentiator";
	local	functab,	# table of function derivatives
		ruletab,	# table of rules for expressions
		diffProduct,	# subroutine for differentiating products
		diffPower,	# subroutine for differentiating powers
		setup;		# initialization for the differentiator
	export	differentiate,	# the differentiator user interface
		addFunc,	# add a new function to the database
		addRule,	# add a new rule to the database
		rule;		# return the rule for a given type
	
	functab := table();
	ruletab := table();

	# Add the derivative of a function (by name) to the
	# internal database.
	addFunc := proc( fname::name, fderiv )
		functab[ fname ] := fderiv
	end proc;

	# Add a new rule, or rules, to the internal database.
	# The type T must be a surface type, as returned by
	# the procedure `whattype()'. The implementation of
	# the rule `impl' is an expression for the derivative
	# of expressions of type T.
	addRule := proc( T, impl )
		if type( T, '{ set, list }' ) then
			map( procname, args )
		elif type( T, 'And( symbol, type )' ) then
			if T = ':-function' then
				error "cannot assign a rule for expressions of type function"
			end if;
			if assigned( ruletab[ T ] ) then
				WARNING( sprintf( "overwriting rule for "
				     "derivatives of type %s", T ) )
			end if;
			ruletab[ T ] := impl
		else
			error "expecting a type name, but got %1", T
		end if
	end proc;

	# This procedure allows the user to query the rule used
	# for expressions of a given type T.
	rule := proc( T )
		if type( T, 'And( name, type )' ) then
			if assigned( ruletab[ T ] ) then
				eval( ruletab[ T ], 1 )
			else
				error "no rule for expressions of type %1", T
			end if
		else
			error "expecting a type symbol, but got %1", T
		end if
	end proc;

	# This is the main user entry point to the differentiator.
	# It is called with two arguments. The first is the expression
	# whose derivative is to be computed. The second is the variable
	# with respect to which to differentiate the first argument.
	differentiate := proc( expr, var::symbol )
		description "compute the derivative of a function of a single variable";
		local	a, b, e;

		if not has( expr, var ) then
			0
		elif expr = var then
			1
		elif type( expr, 'function' ) and nops( expr ) = 1 then
			# Differentiate a function call. This does not
			# handle every possibility.
			e := op( 0, expr ); # the function
			a := op( 1, expr ); # its single argument

			if type( e, 'name' ) and assigned( functab[ e ] ) then
				functab[ e ]( a ) * procname( a, var )
			elif type( e, '`@`' ) then
				# rewrite: (f@g)(x) --> f(g(x))
				procname( op( 1, e )( op( 2, e )( a ) ), var )
			elif type( e, '`^`' ) then
				# rewrite: (f^g)(x) --> f(x)^g(x)
				procname( op( 1, e )( a ) ^ op( 2, e )( a ), var )
			elif has( a, var ) then
				# "generic" chain rule
				':-derivative'( e )( a ) * procname( a, var )
			else
				'procname( args )'
			end if
		else
			b := whattype( expr );
			if assigned( ruletab[ b ] ) then
				ruletab[ b ]( expr, var )
			else
				'procname( args )'
			end if
		end if
	end proc;

	#################
	# Install Rules #
	#################

	# self-map over collections
	addRule( '{list,set,tabular}', () -> map( differentiate, args ) );

	# additivity
	addRule( '`+`', () -> map( differentiate, args ) );

	# generalized product rule
	diffProduct := proc( expr, var )
		local	i;
		Assert( type( expr, '`*`' ) );
		#add( subsop( i = differentiate( op( i, expr ), var ), expr ),
		#	i = 1 .. nops( expr ) );
		add( differentiate( op( i, expr ), var ) * subsop( i = 1, expr ),
			i = 1 .. nops( expr ) )
	end proc;
	addRule( '`*`', eval( diffProduct ) );

	# This subroutine is used to compute the derivative of a
	# general power expression.
	diffPower := proc( expr, var )
		local	b, e;
		Assert( type( expr, '`^`' ) );
		b, e := op( expr );
		if has( e, var ) then
			expr * ( differentiate( e, var ) * ln( b )
				+ e * differentiate( b, var ) / b )
		else # simpler formula; just use the chain rule
			e * b^(e - 1) * differentiate( b, var )
		end if;
	end proc;
	addRule( '`^`', eval( diffPower ) );

	#####################
	# Install Functions #
	#####################

	addFunc( 'exp', exp );
	addFunc( 'ln', x -> 1/x );

	addFunc( 'sin', cos );
	addFunc( 'cos', x -> -sin(x) );
	addFunc( 'tan', x -> 1 + tan(x)^2 );
	addFunc( 'csc', x -> -csc( x ) * cot( x ) );
	addFunc( 'sec', x -> sec( x ) * tan( x ) );
	addFunc( 'cot', x -> -1 - cot( x )^2 );

	addFunc( 'arcsin', x -> 1 / sqrt( 1 - x^2 ) );
	addFunc( 'arccos', x -> -1 / sqrt( 1 - x^2 ) );
	addFunc( 'arctan', x -> 1 / ( 1 + x^2 ) );
	addFunc( 'arccsc', x -> -1 / ( x^2 * sqrt( 1 - 1 / x^2 ) ) );
	addFunc( 'arcsec', x -> 1 / ( x^2 * sqrt( 1 - 1 / x^2 ) ) );
	addFunc( 'arccot', x -> -1 / ( 1 + x^2 ) );

	addFunc( 'sinh', cosh );
	addFunc( 'cosh', sinh );
	addFunc( 'tanh', x -> 1 - tanh( x )^2 );
	addFunc( 'csch', x -> -csch( x ) * coth( x ) );
	addFunc( 'sech', x -> -sech( x ) * tanh( x ) );
	addFunc( 'coth', x -> 1 - coth( x )^2 );

	addFunc( 'arcsinh', x -> 1 / sqrt( 1 + x^2 ) );
	addFunc( 'arccosh', x -> 1 / ( sqrt( ( 1 - x ) ) *  sqrt( ( 1 + x ) ) ) );
	addFunc( 'arctanh', x -> 1 / ( 1 - x^2 ) );
	addFunc( 'arccsch', x -> -1 / ( x^2 * sqrt( 1 + 1 / x^2 ) ) );
	addFunc( 'arcsech', x -> -1 / ( x^2 * sqrt( 1 / x - 1 ) * sqrt( 1 / x + 1 ) ) );
	addFunc( 'arccoth', x -> -1 / ( x^2 - 1 ) );
	# ... etc.

end module:
#savelib( 'DiffImpl' ):
$ifdef	LOAD_REPOSITORY
savelib( 'DiffImpl' ):
$endif	# LOAD_REPOSITORY

differentiate := DiffImpl:-differentiate:
#savelib( 'differentiate' ):
$ifdef	LOAD_REPOSITORY
savelib( 'differentiate' ):
$endif	# LOAD_REPOSITORY

########################################################################
########################################################################
###
###	End of File: DiffImpl.mpl
###
########################################################################
########################################################################
