########################################################################
########################################################################
###
###	Filename: adjmat.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
########################################################################
########################################################################
###
###	NAME
###
###		adjmat -- compute the adjaceny matrix of a graph
###
###	SYNOPSIS
###
###		adjmat( g::Graph )::Matrix
###
###	PARAMETERS
###
###		g -- a graph object that satisfies the Graph interface
###
###	DESCRIPTION
###
###	This is a generic procedure for computing adjacency matrices of
###	directed graphs that satisfy the `Graph' object interface.
###	The adjacency matrix m of a graph g is the square matrix whose
###	rows and columns are equal in number to the order of the graph
###	g, and which are indexed by the vertices of g, in such a way that
###	the matrix entry (u,v) (where u,v are vertices of g) is equal to
###	1 when there is a directed edge from u to v, and 0 otherwise.
###
###	The algorithm demonstrated here is suitable for sparse graphs.
###
########################################################################
########################################################################

$include	"Edge.mi"
$include	"Graph.mi"

adjmat := proc( g::Graph )
	description "compute the adjaceny matrix of a graph";
	local	n,	# the order of the graph g
		m,	# the adjacency matrix; returned
		E,	# the set of edges of the graph g
		V,	# the set of vertices of the graph g
		e,	# induction variable (edges)
		row,	# the row index of a vertex
		col;	# the column index of a vertex
	# Using the option remember avoids expensive recomputation on
	# the same input graph. We also use option system, since we
	# don't mind having the remember table collected if the system
	# runs short on heap storage. In general, you should always
	# use option system when you use option remember, unless you
	# can prove that you will get incorrect results otherwise.
	option	remember, system;

	# Determine the number of vertices
	n := Order( g );
	# Initialize the adjacency matrix
	m := Matrix( n, n, 'storage' = 'sparse' );
	# Determine the edge set
	E := Edges( g );
	# Determine the vertex set. Sorting the list makes this
	# routine deterministic within each Maple session.
	V := sort( convert( Vertices( g ), 'list' ) );
	# Iterate over the edge set
	for e in E do
		# Use `member()' to determine the position of the source
		# and target vertices of the edge in the ordered list of
		# vertices.
		if not member( Source( e ), V, 'row' )
		    or not member( Target( e ), V, 'col' ) then
		    	# This is just ``defensive programming''; this
			# is not supposed to happen.
			error "inconsistent graph structure"
		end if;
		Assert( 1 <= row and row <= n );
		Assert( 1 <= col and col <= n );
		m[ row, col ] := 1
	end do;
	# Return the adjacency matrix
	m
end proc:
#savelib( 'adjmat' ):
$ifdef	LOAD_REPOSITORY
savelib( 'adjmat' ):
$endif	# LOAD_REPOSITORY

########################################################################
########################################################################
###
###	End of File: adjmat.mpl
###
########################################################################
########################################################################
