########################################################################
########################################################################
###
###	Filename: segment.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
########################################################################
########################################################################
###
###	Description: This file defines the `segment' submodule. Unlike
###	the `point' submodule, which can be used independently of the
###	`Shapes' package as a top-level module, this submodule makes
###	reference to lexically scoped variables from the parent module
###	`:-Shapes:-Shapes' (namely, the `point' submodule).
###
########################################################################
########################################################################

# These macros indicate the operand of the segment data structure
# used to represent the start and end ponts of the segment.
macro( START = 1 ):
macro( END = 2 ):

segment := module()
	description "subroutines for segment shapes";
	export	make,		# segment constructor
		area,		# area subroutine for segments
		len,		# compute the length of a segment
		midpoint,	# compute the midpoint of a segment
		circumference,	# compute the circumference of a segment
		start_point,	# extract the start point of a segment
		end_point;	# extract the end point of a segment

	# This is the constructor for segments. A segment is represented
	# by an unevaluated function call with arguments that represent
	# the start and end points of the segment. The points are expected
	# to be points as constructed by this package, though that is not
	# checked here.
	make := ( startPt, endPt ) -> 'SEGMENT'( startPt, endPt );

	# These are two syntax procedures for segments. They return,
	# respectively, the start and end points of a segment. Clients
	# are expected to use these procedures, rather than accessing
	# the segment data structure directly, via `op()'. This allows
	# the representation of segments to be changed in the future.
	start_point := seg -> op( START, seg );
	end_point := seg -> op( END, seg );

	# Compute the length of a segment.
	len := proc( seg )
		local	a, b,		# start and end points of SEG
			a1, b1,		# "x" coordinates of a and b
			a2, b2;		# "y" coordinates of a and b

		# Since the representation of segments is "known" within
		# this module, we can use direct calls to `op()' for
		# improved performance.
		a := op( START, seg );
		b := op( END, seg );

		# Use the syntax procedures provided by the `point' module
		# to extract the "x" and "y" coordinates of the two points.
		# The `xcoord()' and `ycoord()' procedures are visible within
		# this submodule (they belong to the `point' submodule) only
		# through a partially qualified reference. That is, we cannot
		# directly call them as `xcoord()' and `ycoord()'.
		a1 := point:-xcoord( a );
		a2 := point:-ycoord( a );
		b1 := point:-xcoord( b );
		b2 := point:-ycoord( b );

		# Compute the length using the data gathered above.
		sqrt( ( b2 - a2 )^2 + ( b1 - a1 )^2 )
	end proc;

	# Compute the midpoint of a segment.
	midpoint := proc( seg )
		local	a, b,		# start and end points of SEG
			a1, b1,		# "x" coordinates of a and b
			a2, b2;		# "y" coordinates of a and b

		# Extract the endpoints of the segment. In this case,
		# the syntax procedures defined for segments are used.
		# The advantage of this approach is that this procedure
		# will likely continue to work without modification even
		# if the representation of segments change. In that case,
		# only the syntax procedures will require modification.
		# This approach is preferred unless this particular usage
		# turns out to be a performance bottleneck (as revealed
		# by profiling the code).
		a := start_point( seg );
		b := end_point( seg );

		# Extract the "x" and "y" coordinates of the end points.
		# Four local variables could be saved by "inlining" these
		# calls in the call to the point constructor below. Since
		# the performance impact of this alternative is so tiny,
		# the improved readability of the approach taken here is
		# deemed more important.
		a1 := point:-xcoord( a );
		a2 := point:-ycoord( a );
		b1 := point:-xcoord( b );
		b2 := point:-ycoord( b );

		# Call the point constructor to generate the midpoint
		# of the segment from its coordinates, which are
		# computed by using the data gathered above.
		point:-make( ( a1 + b1 ) / 2, ( a2 + b2 ) / 2 );
	end proc;

	# Just as for points, the area and circumference computations
	# are trivial for segments.
	area := () -> 0;
	circumference := () -> 0;

end module:

########################################################################
########################################################################
###
###	End of File: segment.mpl
###
########################################################################
########################################################################
