########################################################################
########################################################################
###
###	Filename: shapes.mpl
###
###     Copyright (c) 1999 Waterloo Maple, Inc.
###
###     THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF WMI
###     The copyright notice above does not evidence any
###     actual or intended publication of such source code.
###
########################################################################
########################################################################
###
###	Description: This is the "master" source file for the SHAPES
###	package. The SHAPES package is a simple demonstration package.
###	You can use this package source as a template for your own
###	packages, by copying the source files and replacing the Maple
###	code with your own package code.
###
###	The purpose of this package is merely to illustrate some proven
###	practices for structuring a Maple package. The sample SHAPES
###	package has very little of mathematical interest. It is
###	intended to illustrate the software engineering principles
###	behind packages, and to provide a few coding techniques that
###	are helpful in structuring a Maple package.
###
###	The SHAPES package is implemented as a module. This is now the
###	recommended implementation for packages. Although the SHAPES
###	package is itself too small to warrant it, the package is also
###	organized into multiple source files, and the module structure
###	has been designed hierarchically to demonstrate a pattern of
###	code management that scales well to larger packages.
###
###	The code is extensively commented. The intent is to explain
###	what each aspect of the code is doing and why.
###
########################################################################
########################################################################

# This is a preprocessor directive that requests the
# standard include file `assert.mi' be included at this
# point. This include file simply defines two preprocessor
# macros `Assert()' and `Assert2()' (two-argument form),
# which expand to calls to the built-in procedure `ASSERT()' if
# the preprocessor symbol `_DEBUG_' is defined, and which
# otherwise expands to nothing.
$include	<assert.mi>

# The macro `ALL_SHAPES' defines the set of all shape names
# understood by this package.
macro( ALL_SHAPES = {
	'point',
	'segment',
	'circle',
	'square',
	'triangle'
} ):

# This statement defines the main package module, by assigning a module
# expression to the global name `Shapes'.
Shapes := module()

	# A package should have a description string, which is printed
	# as part of the pretty-printed form of the instantiated module.
	description "a package for computing with simple "
			"geometric shapes";

	# Use the EXPORT statement to list the names of the top-level
	# routines provided by the package. This simple package provides
	# only three routines: `make', `area', and `circumference'. These
	# names are the ones made available in short form when the user
	# applies the `with()' procedure to this package.
	export	make,
		area,
		circumference;
	# In addition to the user-visible, exported names listed in the
	# EXPORT statement, a module-based package may also have data
	# that is not user-accessible. These variables are declared in the
	# LOCAL statement. The SHAPES package has two local variables:
	# `Shapes' and `circum_table'.  The `Shapes' variable is assigned
	# a submodule, and it is used for dispatching on the shape being
	# manipulated. The local variable `circum_table' is assigned a
	# dispatch table, which is another method dispatch mechanism.
	local	Shapes,
		circum_table;

	# Here, the `Shapes' submodule is defined. It has the same name
	# as its parent module, so its fully qualified name is :-Shapes:-Shapes.
	# Normally, to avoid confusion, a different name would likely
	# have been chosen. The name `Shapes' was chosen for this example
	# only to illustrate the point that this is possible. The name
	# `Shapes', in the body of this module definition, refers to the
	# local variable `Shapes'. (The global name `Shapes' is always
	# accessible via the syntax `:-Shapes'.)
	Shapes := module()
		# Since this submodule is local to the parent module, and
		# not accessible to users, we did not bother to attach a
		# description string. (It does not hurt to add one if
		# you want.)

		# The exports of this submodule are all themselves submodules.
		# Here, a distinct submodule is defined for each kind of
		# shape supported by the package. The choice of names is
		# significant in the implementation of the dispatch mechanism for
		# the `make' export of the main :-Shape package.
		export	point,
			segment,
			circle,
			square,
			triangle;


# To illustrate the technique, this package has been subdivided into
# several source code files. Each of the exported submodules of the
# module :-Shape:-Shape is defined in its own source file. This
# subdivision reflects the logical structure of the package, and
# facilitates modifications to add new shapes. An important benefit
# of this approach is that it facilitates team development; it is easy
# for different developers to work on different shapes.
$include	"point.mpl"
$include	"segment.mpl"
$include	"circle.mpl"
$include	"square.mpl"
$include	"triangle.mpl"
	end module;

	# This is the exported procedure `make'. It is called to construct
	# one of the "shape" objects supported by the package. The kind
	# of shape is specified by a name passed as the first argument.
	# The name of the shape must match that of the exported submodule
	# of the :-Shapes:-Shapes submodule. Extra arguments are used to
	# specify the particular parameters required for each kind of
	# shape. Since the necessary arguments, and their meanings, differ
	# after the first (shape name) argument, any further argument-
	# checking is left to the specific constructors.
	make := proc( what::symbol )
		description "constructor for shapes";
		local	ctor,		# the shape constructor, if found
			theShape;	# the submodule for the kind of shape requested

		# Query the :-Shapes:-Shapes submodule for the appropriate
		# kind of shape. If it is not found, issue an error explaining
		# what went wrong.
		if not member( what, Shapes, 'theShape' ) then
			error "shape `%1' not available", what
		end if;
		if member( ':-make', theShape, 'ctor' ) then
			# The local variable `ctor' has been assigned, as a
			# side effect of the call to `member()', the local
			# instance of the `make' procedure in the correct
			# shape-specific submodule.
			ctor( args[ 2 .. nargs ] )
		else
			# All of the shape-specific submodules export a `make'
			# procedure, so this case should never be entered.
			# Defensive programming, however, encourages a useful
			# error message, and is robust against future changes.
			error "no constructor provided for shape %1", what
		end if
	end proc;

	# This is the exported procedure `area()', which returns
	# the area of a given shape (returned by `make()'). Since
	# each shape is represented by an unevaluated function call,
	# the 0th operand of which is a mnemonic for the shape type,
	# this procedure can be used as a kind of "tag" on which
	# dispatching to the appropriate shape-specific subroutine
	# can be made. This procedure uses a straightforward conditional
	# type of dispatching. This method is suitable for small to
	# medium numbers of cases.
	area := proc( shape )
		description "compute the area of a shape";
		local	tag;

		if not type( shape, 'function' ) then
			# This is a minimal (and fast) type check that is
			# done only to ensure that this procedure can be
			# executed correctly.
			error "expecting a shape expression, but got %1", shape
		end if;

		# Extract the "tag" information from the shape
		tag := op( 0, shape );

		# Select the `area()' subroutine in the appropriate
		# submodule based on the `tag' value. Pass the shape
		# expression along so the shape-specific `area()' routine
		# can extract from it the necessary data to compute
		# the area.
		if tag = ':-POINT' then
			Shapes:-point:-area( shape )
		elif tag = ':-SEGMENT' then
			Shapes:-segment:-area( shape )
		elif tag = ':-CIRCLE' then
			Shapes:-circle:-area( shape )
		elif tag = ':-SQUARE' then
			Shapes:-square:-area( shape )
		elif tag = ':-TRIANGLE' then
			Shapes:-triangle:-area( shape )
		else
			# Even though this case is not expected to be
			# reached, be careful to issue a useful diagnostic
			# to guard against programming errors and future
			# modifications to the package.
			error "not a recognized shape: %1", tag
		end if
	end proc;

	# The local variable `circum_table' is assigned a table of
	# procedures used to dispatch to the appropriate shape-specific
	# circumference routine. This technique scales well to many
	# cases. It is the fastest of the three dispatch techniques
	# illstrated here, but requires storage for the dispatch table.
	# As table indices, the shape "tags" are used again, since they
 	# can be extracted from the input shape quite readily.
	circum_table := table();
	circum_table[ 'POINT' ] := Shapes:-point:-circumference;
	circum_table[ 'SEGMENT' ] := Shapes:-segment:-circumference;
	circum_table[ 'CIRCLE' ] := Shapes:-circle:-circumference;
	circum_table[ 'SQUARE' ] := Shapes:-square:-circumference;
	circum_table[ 'TRIANGLE' ] := Shapes:-triangle:-circumference;

	# This is the exported procedure `circumference'. It takes a
	# shape expression as an argument, and returns the computed
	# circumference. The table `circum_table' defined above is
	# used for dispatching to the appropriate subroutine.
	circumference := proc( shape )
		description "compute the circumference of a shape expression";
		if not type( shape, 'function' ) then
			# In keeping with the general design, only the
			# superficial type check for an expression of
			# type `function' is done here, to ensure that
			# this routine will not fail in unexpected ways.
			# Checking the well-formedness of the shape
			# object is left to the shape-specific subroutines.
			error "expecting a shape, but got %1", shape
		end if;
		if assigned( circum_table[ op( 0, shape ) ] ) then
			circum_table[ op( 0, shape ) ]( shape )
		else
			error "no circumference method is available for shape %1."
				" Supported shapes are: %2",
				tag, sprintf( "%q", op( ALL_SHAPES ) )
		end if
	end proc;

end module: # Shape

# Now that the package has been defined and is assigned to the package name
# `Shapes', it has to be saved to a repository. The assigned values
# of the global variables `savelibname' and `libname' are used to
# determine which Maple repository the package will be saved in when
# a call to `savelib()' is executed.

# We illustratre two methods of coding a way to save the package to
# the repository. Both rely on the environment variable `MAPLELIB' 
# being set at the time this file is loaded. If you are working on 
# a system that does not provide this facility, you can use
# command-line options to the command-line version of Maple, or
# a hardcoded value of `savelibname' in this file. For example:
# savelibname := "C:\\Maple\\MyProjects": # DOS/Windows
# savelibname := cat( getenv("HOME"), "/maple/lib" ): # UNIX
# savelibname := "Macintosh HD:Maple:My Projects": # Mac
# Another alternative is to set the global `savelibname' with the
# preprocessor. This can be effected by adding a line such as the
# following.
# savelibname := SAVELIBNAME:
# or by defining it conditionally as in
# $ifdef SAVELIBNAME
# savelibname := SAVELIBNAME:
# $endif
# Then load this file by including the command-line argument
# "-D <directory where your repository is>" to Maple. An example command
# line for UNIX, for instance, might look like:
# maple -s -q -D $HOME/maple/lib -D LOAD_REPOSITORY shapes.mpl

# Method 1: Use a *commented* call to `savelib()'. The shell script
# `mload', or a makefile rule, can be used to uncomment this line during
# loading of the file, which causes the call to `savelib()' to be
# evaluated. This allows the file to be processed without loading
# the package as well.
#savelib( 'Shapes' ):

# Method 2: Use the preprocessor to conditionally execute the call
# to `savelib()'. This can be done in a script or a makefile rule.
$ifdef	LOAD_REPOSITORY
savelib( 'Shapes' ):
$endif	# LOAD_REPOSITORY

# NOTE 1: In either case, *only* the name `Shapes' has to be saved.
# Since everything in the package is a subexpression of the module
# Shapes, the entire package is saved in this one call.

# NOTE 2: The arguments to `savelib()' (which must all be names) should
# always be quoted using unevaluation quotes. This ensures that the
# value passed to `savelib()' (after argument evaluation) is the name
# and not its assigned value. Often, last name evaluation rules allow
# you to get away with not quoting the saved names, but defensive
# programming dictates that you should do otherwise.

########################################################################
########################################################################
###
###	End of File: shapes.mpl
###
########################################################################
########################################################################
